@@ -228,6 +228,9 @@ manager_select_theme_for_job (GSManager *manager,
                 info = gs_theme_manager_lookup_theme_info (manager->priv->theme_manager, theme);
                 if (info != NULL) {
                         command = gs_theme_info_get_exec (info);
+                } else {
+                        gs_debug ("Could not find information for theme: %s",
+                                  theme);
                 }
 
                 gs_job_set_command (job, command);
@@ -1155,7 +1158,7 @@ static void
 add_unfade_idle (GSManager *manager)
 {
         remove_unfade_idle (manager);
-        manager->priv->unfade_idle_id = g_idle_add ((GSourceFunc)unfade_idle, manager);
+        manager->priv->unfade_idle_id = g_timeout_add (500, (GSourceFunc)unfade_idle, manager);
 }
 
 static gboolean
@@ -1566,3 +1569,12 @@ gs_manager_request_unlock (GSManager *manager)
 
         return TRUE;
 }
+
+void
+gs_manager_cancel_unlock_request (GSManager *manager)
+{
+        GSList *l;
+        for (l = manager->priv->windows; l; l = l->next) {
+                gs_window_cancel_unlock_request (l->data);
+        }
+}
